/*
* Copyright 1993-2006 NVIDIA Corporation.  All rights reserved.
*
* NOTICE TO USER:
*
* This source code is subject to NVIDIA ownership rights under U.S. and
* international Copyright laws.
*
* NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
* CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
* IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
* REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
* IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
* OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
* OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
* OR PERFORMANCE OF THIS SOURCE CODE.
*
* U.S. Government End Users.  This source code is a "commercial item" as
* that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of
* "commercial computer software" and "commercial computer software
* documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995)
* and is provided to the U.S. Government only as a commercial end item.
* Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
* 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
* source code with only those rights set forth herein.
*/

/*
* This sample demonstrates how use texture fetches in CUDA
*
* This sample takes an input PGM image (image_filename) and generates 
* an output PGM image (image_filename_out).  This CUDA kernel performs
* a Hough transform on the texture using global atomics for locking
* output memory (compute capability 1.1 required!)
*/

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include <cutil.h>

// includes, kernels
#include <AtomicHough_kernel.cu>

char *image_filename = "polygon_bw.pgm";

////////////////////////////////////////////////////////////////////////////////
// declaration, forward
void runTest( int argc, char** argv);

extern "C"
void computeGold( float *g_idata, float* g_odata, int iwidth, int iheight, int owidth, int oheight, float max_rho, unsigned int shared_floats, float scaling );

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int
main( int argc, char** argv) 
{
    runTest( argc, argv);

    CUT_EXIT(argc, argv);
}

////////////////////////////////////////////////////////////////////////////////
//! Run a simple test for CUDA
////////////////////////////////////////////////////////////////////////////////
void
runTest( int argc, char** argv) 
{
    CUT_DEVICE_INIT(argc, argv);
//    CUT_DEVICE_INIT();
  
    // load image from disk
    float* h_input = NULL;
    unsigned int input_width, input_height;
    char* image_path = cutFindFilePath(image_filename, argv[0]);
    if (image_path == 0)
        exit(EXIT_FAILURE);
    CUT_SAFE_CALL( cutLoadPGMf(image_path, &h_input, &input_width, &input_height));

    unsigned int input_size = input_width * input_height * sizeof(float);
    printf("Loaded '%s', %d x %d pixels\n", image_filename, input_width, input_height);
    
    float max_rho = sqrt((float)input_width*input_width+input_height*input_height)/2;
    printf("Max_Rho: %f (pixels)\n", max_rho);

    // define output dimensions
    unsigned int output_width = 512;  // angular resolution [0...PI[
    unsigned int output_height = (int)max_rho; // 512;  // distance resolution [0...max_rho]

    // Block Size
    dim3 dimBlock(THETA_THREADS, HORIZONTAL_GROUPING, 1);

    // Grid Size: Image width and height must be a multiple of HORIZONTAL_GROUPING and VERTICAL_GROUPING
    dim3 dimGrid((input_width+HORIZONTAL_GROUPING-1) / HORIZONTAL_GROUPING, (input_height+VERTICAL_GROUPING-1) / VERTICAL_GROUPING, 1);
    int numblocks = dimGrid.x * dimGrid.y;

    printf("Threads per Block: %d\n", dimBlock.x * dimBlock.y * dimBlock.z);
    printf("Blocks within Grid: %d\n", numblocks);

    // memory requirements for final result array
    unsigned int result_size = output_width * output_height * sizeof(float);

    // allocate and clear device memory for result
    float* d_result = NULL;
    CUDA_SAFE_CALL( cudaMalloc( (void**) &d_result, result_size));
    CUDA_SAFE_CALL( cudaMemset( d_result, 0, result_size));

    // allocate array and copy image data
    cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
    cudaArray* cu_array;
    CUDA_SAFE_CALL( cudaMallocArray( &cu_array, &channelDesc, input_width, input_height )); 
    CUDA_SAFE_CALL( cudaMemcpyToArray( cu_array, 0, 0, h_input, input_size, cudaMemcpyHostToDevice));

    // set texture parameters
    tex.addressMode[0] = cudaAddressModeClamp;
    tex.addressMode[1] = cudaAddressModeClamp;
    tex.filterMode = cudaFilterModeLinear; // cudaFilterModePoint;
    tex.normalized = false;    // access with non-normalized texture coordinates

    // Bind the array to the texture
    CUDA_SAFE_CALL( cudaBindTextureToArray( tex, cu_array, channelDesc));

    // Compute size of shared memory block and corresponding section_height variable
    int shared_size;
    int section_height = MIN(output_height, MAX_SECTION_HEIGHT);
    do {
        shared_size = THETA_THREADS * section_height * sizeof(float);
    } while (shared_size > 16000 - MAX_SECTION_HEIGHT*sizeof(unsigned char) && --section_height);
    
    printf("Shared size: %d (bytes)\n", shared_size);
    printf("Section height: %d\n", section_height);
    
    // allocate and clear device memory for mutexes
    unsigned int num_mutex = (section_height + HORIZONTAL_GROUPING-1) / HORIZONTAL_GROUPING;
    printf("Number of mutexes: %d\n", num_mutex);
    size_t mutex_size = num_mutex * sizeof(int);  // 256 mutexes should be enough for everybody
    int* d_mutex = NULL;
    CUDA_SAFE_CALL( cudaMalloc( (void**) &d_mutex, mutex_size));
    CUDA_SAFE_CALL( cudaMemset( d_mutex, 0, mutex_size));

    // Number of Accumulation Rounds: output_width must be a multiple of THETA_THREADS!
    int theta_rounds = (output_width + THETA_THREADS-1) / THETA_THREADS;
    printf("Number of accumulation rounds per thread: %d\n", theta_rounds);

    // This factor adjusts brightness of resulting image
    float scaling = 1.0/16;  // downscaling of resulting brightness levels
    float limit = 1.0; // to prevent brightness overflow/wraparound when writing PGM.
                       // NOTE: do not limit when you intend to analyze the output data

    // grid and block dimensions for final accumulation
    dim3 accuGrid((output_width+HORIZONTAL_GROUPING-1) / HORIZONTAL_GROUPING, 1, 1);
    dim3 accuBlock(HORIZONTAL_GROUPING, 8, 1); // yields 64 * 8 = 512 threads per block

    #if !defined(__DEVICE_EMULATION__)
    // warmup
    transformKernel<<< dimGrid, dimBlock, shared_size >>>( d_result, d_mutex, input_width, input_height, output_width, output_height, max_rho, section_height, theta_rounds, shared_size/sizeof(float), scaling, limit );
    // check if kernel execution generated an error
    CUT_CHECK_ERROR("Kernel execution failed");
    CUDA_SAFE_CALL( cudaMemset( d_result, 0, result_size));
    #endif

    CUDA_SAFE_CALL( cudaThreadSynchronize() );
    unsigned int timer = 0;
    CUT_SAFE_CALL( cutCreateTimer( &timer));
    CUT_SAFE_CALL( cutStartTimer( timer));

    // execute the kernels
    transformKernel<<< dimGrid, dimBlock, shared_size >>>( d_result, d_mutex, input_width, input_height, output_width, output_height, max_rho, section_height, theta_rounds, shared_size/sizeof(float), scaling, limit );
    // check if kernel execution generated an error
    CUT_CHECK_ERROR("Kernel execution failed");

    CUDA_SAFE_CALL( cudaThreadSynchronize() );
    CUT_SAFE_CALL( cutStopTimer( timer));
    printf("Processing time: %f (ms)\n", cutGetTimerValue( timer));
    printf("%.2f Mpixels/sec\n", (input_width*input_height / (cutGetTimerValue( timer) / 1000.0f)) / 1e6);
    CUT_SAFE_CALL( cutDeleteTimer( timer));

    // allocate mem for the result on host side
    float* h_result = (float*) malloc( result_size );
    // copy result from device to host
    CUDA_SAFE_CALL( cudaMemcpy( h_result, d_result, result_size, cudaMemcpyDeviceToHost) );

    // write result to file
    char result_filename[1024];
    sprintf(result_filename, "%s_out.pgm", image_path);
    CUT_SAFE_CALL( cutSavePGMf( result_filename, h_result, output_width, output_height));
    printf("Wrote '%s', %d x %d pixels\n", result_filename, output_width, output_height);

    // write regression file if necessary
    if( cutCheckCmdLineFlag( argc, (const char**) argv, "regression")) 
    {
        // write file for regression test
        CUT_SAFE_CALL( cutWriteFilef( "./data/regression.dat", h_result, output_width*output_height, 0.0));
    }

    // cleanup memory
    CUDA_SAFE_CALL(cudaFree(d_mutex));
    CUDA_SAFE_CALL(cudaFree(d_result));
    CUDA_SAFE_CALL(cudaFreeArray(cu_array));
    free(h_result);
    cutFree(h_input);
    cutFree(image_path);
}
