/*
 * Copyright 1993-2006 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:   
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and 
 * international Copyright laws.  
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE 
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR 
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH 
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF 
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.   
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL, 
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS 
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE 
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE 
 * OR PERFORMANCE OF THIS SOURCE CODE.  
 *
 * U.S. Government End Users.  This source code is a "commercial item" as 
 * that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of 
 * "commercial computer software" and "commercial computer software 
 * documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995) 
 * and is provided to the U.S. Government only as a commercial end item.  
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through 
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the 
 * source code with only those rights set forth herein.
 */

#define PI 3.1415926535897932384626433832795f

#ifndef _ATOMICHOUGH_KERNEL_H_
#define _ATOMICHOUGH_KERNEL_H_

#include <sm_11_atomic_functions.h>

// Threads per block is defined here. Cannot currently reach 512 due to high register user per thread.
// THETA_THREADS is blockDim.x and HORIZONTAL_GROUPING is blockDim.y.
// working combinations are 8/48, 16/24, 32/12, all resulting in 384 threads
#define THETA_THREADS       8   // Number of theta values that each block processes simultaneously.
#define HORIZONTAL_GROUPING 48  // Stripe width in pixels that each block processes.

// maximum height of a rho section that can be computed in one pass over theta
#define MAX_SECTION_HEIGHT 512

// Stripe height that each block processes. I guess we could also pass this as a kernel argument
// and always make this the input image height
#define VERTICAL_GROUPING   512

//#define SDATA(index)      CUT_BANK_CHECKER(sdata, index)
#define SDATA(index)        sdata[index]

// simple min and max macros
#define MAX(A,B) ((A)>(B)?(A):(B))
#define MIN(A,B) ((A)<(B)?(A):(B))

// declare texture reference for 2D float texture
texture<float, 2, cudaReadModeElementType> tex;


__device__ inline void atomicFloatAdd(float *address, float val)
{
      volatile int i_val = __float_as_int(val);
      volatile int tmp0 = 0;
      volatile int tmp1;

      while( (tmp1 = atomicCAS((int *)address, tmp0, i_val)) != tmp0)
      {
              tmp0 = tmp1;
              i_val = __float_as_int(val + __int_as_float(tmp1));
      }
}


////////////////////////////////////////////////////////////////////////////////
//! Hough transform an image using texture lookups
//! This kernel generates blockwise intermediate results
//! @param g_odata  output in global memory
//! @param g_mutex  array of integer mutexes in global memory
//! @param iwidth   input image width
//! @param iheight  input image height
//! @param owidth   result width (theta dimension)
//! @param oheight  result height (rho dimension)
//! @param section_height  the height of a rho section fitting in shared memory
//! @param rounds   number of accumulation roudns required to cover the theta axis
//! @param shared_floats  number of floats in sdata, used for bounds checking
////////////////////////////////////////////////////////////////////////////////
__global__ void
transformKernel( float* g_odata, int *g_mutex, int iwidth, int iheight, int owidth, int oheight, float max_rho, int section_height, int rounds, unsigned int shared_floats, float scaling, float limit ) 
{
    extern __shared__ float sdata[];

    // Used to prevent collisions during accumulation in shared memory.
    volatile unsigned char __shared__ anticollider[MAX_SECTION_HEIGHT];

    // calculate normalized texture coordinates
    volatile int i_x = blockIdx.x * HORIZONTAL_GROUPING + threadIdx.y;
    volatile int i_y = blockIdx.y * VERTICAL_GROUPING;
    
    // image-centered coordinates
    volatile float x = (float)i_x - (float)iwidth / 2  + 0.5;
    volatile float y = (float)i_y - (float)iheight / 2 + 0.5;

    // compute range of theta that we cover in each iteration (round)
    volatile float thetaoffset = (2*PI * (rounds * THETA_THREADS) / owidth) / rounds;

    // perform multiple accumulation rounds, each processing a limited range of theta only
    for (int theta_iter = 0; theta_iter < rounds; theta_iter++)
    {
        // due to restricted size of shared memory we have to make several passes along the rho axis
        for (int section_start = 0; section_start < oheight; section_start += section_height)
        {
            // Before each iteration, clear accumulators in shared memory. This should not have any bank conflicts.
            // we do this in several "stripes" indexed by rho_iter
            for (int rho_iter = 0; rho_iter < (section_height + HORIZONTAL_GROUPING-1) / HORIZONTAL_GROUPING; rho_iter++)
            {
                volatile int ypos = rho_iter * HORIZONTAL_GROUPING + threadIdx.y;
                
                // grid is based on input_width, hence this check is needed to make sure we're within bounds
                if (ypos < section_height)
                {
                    #if defined(__DEVICE_EMULATION__)
                    if ( (rho_iter * HORIZONTAL_GROUPING + threadIdx.y) * THETA_THREADS + threadIdx.x > shared_floats )
                        printf("Out of bounds (clear) (%d>%d)!\n", (rho_iter * HORIZONTAL_GROUPING + threadIdx.y) * THETA_THREADS + threadIdx.x, shared_floats);
                    #endif
                    SDATA( ypos * THETA_THREADS + threadIdx.x ) = 0;
                }
            }

            __syncthreads();
             
            // compute the angle theta that is thread-dependent
            volatile float theta = thetaoffset * ((float)theta_iter + (float)threadIdx.x / THETA_THREADS) - PI;

            // process several pixels along the Y axis within each thread.
            // TODO: This loop has potential to be (partially) unrolled
            for (int yoff = 0; yoff < VERTICAL_GROUPING; yoff++)
            {
                // bounds check such that iwidth does not have to be a multiple of HORIZONTAL_GROUPING
                // and owidth does not have to be a multiple of THETA_THREADS
                if (theta < PI && i_x < iwidth)
                {
                    // bounds check such iheight does not have to be a multiple of VERTICAL_GROUPING
                    if ( i_y+yoff < iheight )
                    {
                        // compute the distance rho from center point
                        volatile float rho = x * __cosf(theta) + (y+(float)yoff) * __sinf(theta);

                        // find vertical index into shared memory accumulator array
                        volatile int s_rho   = (int)(rho * oheight / max_rho) - section_start;
                        
                        // ignore all results that don't fall into the current rho section
                        if (s_rho >= 0 && s_rho < section_height)
                        {
                            // read from texture from a thread-dependent X position (Y coordinate is common)
                            float value = tex2D(tex, (float)i_x, (float)(i_y+yoff));
                            if (value > 0) {
                                volatile int s_index = s_rho * THETA_THREADS + threadIdx.x;
                                
                                // unfortunately we have to split up the warp into several divergent
                                // paths to perform collision-free accumulations in shared memory.
                                volatile int warpthread = ((blockDim.x * threadIdx.y + threadIdx.x) % warpSize)/THETA_THREADS;
                                for (int i=0; i < warpSize / THETA_THREADS; i++) {
                                    while (warpthread == i) {
                                        anticollider[ s_rho ] = threadIdx.y;
                                        if ( anticollider[ s_rho ] == threadIdx.y ) {
                                            SDATA( s_index ) = SDATA( s_index ) + value;
                                            break;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }

            // After each iteration, add accumulators to global memory. Try to coalesce writes for speed.
            // we do this in several "stripes" indexed by rho_iter
            for (int rho_iter = 0; rho_iter < (section_height + HORIZONTAL_GROUPING-1) / HORIZONTAL_GROUPING; rho_iter++)
            {
                volatile int threadnum  = (blockDim.x * threadIdx.y + threadIdx.x);

                volatile int odata_x = theta_iter * THETA_THREADS + threadIdx.x;
                volatile int sdata_y = rho_iter * HORIZONTAL_GROUPING + threadIdx.y;
                volatile int odata_y = section_start + sdata_y;
                
                // offset in global memory
                volatile int ooffset = owidth * odata_y + odata_x;
                
                // offset in shared memory
                volatile int soffset = sdata_y * THETA_THREADS + threadIdx.x;
                
                // lock the section of output memory (indexed by rho_iter) that we're going to write into
                if (threadnum == 0)
                {
                    // first thread in block spins until we've got a lock
                    while (1 == atomicExch(&g_mutex[rho_iter], 1));
                }
                
                __syncthreads();
                
                // grid is based on input_width, hence this check is needed to make sure we're within bounds
                if (sdata_y < section_height && odata_x < owidth && odata_y < oheight)
                {
                    volatile float tmp = g_odata[ ooffset ] + scaling * SDATA ( soffset );
                    if (tmp > limit) tmp = limit;
                    
                    g_odata[ ooffset ] = tmp;
                }

                __syncthreads();

                // lunock the section of output memory (indexed by rho_iter) that we wrote into
                if (threadnum == 0)
                {
                    // unlock mutex
                    atomicExch(&g_mutex[rho_iter], 0);
                }
            }
        }
    }
}

#endif // #ifndef _ATOMICHOUGH_KERNEL_H_
